#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Contains the class to extract results simulated with pyramses."""

import warnings
import os
import errno
from scipy.io import FortranFile
from typing import NamedTuple
import numpy as np
import matplotlib.pyplot as plt

from .globals import RAMSESError, CustomWarning, wrapToList

warnings.showwarning = CustomWarning

def curplot(curves):
    """Plots multiple curves

    :param list curves: the curves to plot

    """
    curves = wrapToList(curves)
    for curve in curves:
        plt.plot(curve.time, curve.value, label=curve.msg)
    plt.legend(loc='best',ncol=2)
    plt.xlabel('time (s)')
    plt.show()

class cur(NamedTuple):
    """Class to save results
    """
    time: np.ndarray
    value: np.ndarray
    msg: str

    def plot(self):
        curplot(self)

class extractor(object):
    """The extractor class is used to extract the timeseries data after the simulation and process them. 
    
    :param str rtraj: the path to the ramses trajectory file. This should be the same file that was set in the case with case.addTrj('filenm.rtrj').

    :Example:

    >>> import pyramses
    >>> case = pyramses.cfg("case.rcfg") # load case from a configuration file
    >>> ram = pyramses.sim()
    >>> ram.execSim(case) # run the simulation
    >>> ext = pyramses.extractor(case.getTrj())
    >>> ext.getBus('1041').mag.plot() # will plot the timeseries simulated for the voltage magnitude on bus '1041'
    """
    
    def __init__(self,traj):
        if not isinstance(traj, str):
            raise TypeError('PyDyngraph: Class Extractor expects a string path of the trajectory file to initialize.')
        
        self._trajfilename = traj
        
        if not os.path.isfile(traj):
            raise FileNotFoundError(errno.ENOENT, 
                                    os.strerror(errno.ENOENT), traj)    
            
        f = FortranFile(traj, 'r')
        
        self._busnum = f.read_ints()[0]
        self._busname = []
        for i in range(self._busnum):
            self._busname.append(f.read_record(dtype="S18")[0].decode('UTF-8').strip())
        
        self._shunum = f.read_ints()[0]
        self._shuname = []
        for i in range(self._shunum):
            self._shuname.append(f.read_record(dtype="S20")[0].decode('UTF-8').strip())
        
        self._ldnum = f.read_ints()[0]
        self._ldname = []
        for i in range(self._ldnum):
            self._ldname.append(f.read_record(dtype="S20")[0].decode('UTF-8').strip())
        
        self._branum = f.read_ints()[0]
        self._braname = []
        for i in range(self._branum):
            self._braname.append(f.read_record(dtype="S20")[0].decode('UTF-8').strip())
        
        self._syncnum = f.read_ints()[0]
        self._syncname = []
        self._excobsnum = []
        self._excobsname = []
        self._adexc = []
        idxexc = 1
        self._torobsnum = []
        self._torobsname = []
        self._adtor = []
        idxtor = 1
        for i in range(self._syncnum):    
            self._syncname.append(f.read_record(dtype="S20")[0].decode('UTF-8').strip())
            self._excobsnum.append(f.read_ints()[0])
            self._excobsname.append([])
            self._adexc.append(idxexc)
            for j in range(self._excobsnum[i]):
                self._excobsname[i].append(f.read_record(dtype="S10")[0].decode('UTF-8').strip())
            idxexc = idxexc + self._excobsnum[i]
            self._torobsnum.append(f.read_ints()[0])
            self._torobsname.append([])
            self._adtor.append(idxtor)
            for j in range(self._torobsnum[i]):
                self._torobsname[i].append(f.read_record(dtype="S10")[0].decode('UTF-8').strip())
            idxtor = idxtor + self._torobsnum[i]
        self._adexc.append(idxexc)
        self._adtor.append(idxtor)
        
        self._injnum = f.read_ints()[0]
        self._injobsnum = []
        self._injobsname = []
        self._injname = []
        self._adinj = []
        idxinj = 1
        for i in range(self._injnum):
            self._adinj.append(idxinj)
            self._injname.append(f.read_record(dtype="S20")[0].decode('UTF-8').strip())
            self._injobsnum.append(f.read_ints()[0])
            self._injobsname.append([])
            for j in range(self._injobsnum[i]):
                self._injobsname[i].append(f.read_record(dtype="S10")[0].decode('UTF-8').strip())
            idxinj = idxinj + self._injobsnum[i]
        self._adinj.append(idxinj)
        
        self._twopnum = f.read_ints()[0]
        self._twopobsnum = []
        self._twopobsname = []
        self._twopname = []
        self._adtwop = []
        idxtwop = 1
        for i in range(self._twopnum):
            self._adtwop.append(idxtwop)
            self._twopname.append(f.read_record(dtype="S20")[0].decode('UTF-8').strip())
            self._twopobsnum.append(f.read_ints()[0])
            self._twopobsname.append([])
            for j in range(self._twopobsnum[i]):
                self._twopobsname[i].append(f.read_record(dtype="S10")[0].decode('UTF-8').strip())
            idxtwop = idxtwop + self._twopobsnum[i]
        self._adtwop.append(idxtwop)
        
        self._dctlnum = f.read_ints()[0]
        self._dctlobsnum = []
        self._dctlobsname = []
        self._dctlname = []
        self._addctl = []
        idxdctl = 1
        for i in range(self._dctlnum):
            self._addctl.append(idxdctl)
            self._dctlname.append(f.read_record(dtype="S20")[0].decode('UTF-8').strip())
            self._dctlobsnum.append(f.read_ints()[0])
            self._dctlobsname.append([])
            for j in range(self._dctlobsnum[i]):
                self._dctlobsname[i].append(f.read_record(dtype="S10")[0].decode('UTF-8').strip())
            idxdctl = idxdctl + self._dctlobsnum[i]
        self._addctl.append(idxdctl)
        
        self._totobs = 2*self._busnum + self._shunum + 2*self._ldnum + 6*self._branum + \
                       13*self._syncnum + sum(self._excobsnum) + sum(self._torobsnum) + \
                       sum(self._injobsnum) + sum(self._twopobsnum) + sum(self._dctlobsnum)
        
        self._results = []
        buffsz = f.read_ints(np.int64)[0]
        while buffsz > 0:
            temp = f.read_reals(dtype=np.float64)
            self._results = np.concatenate((self._results, temp))
            buffsz = f.read_ints(np.int64)[0]
        
        self._results = np.reshape(self._results, (-1,self._totobs+1), order='C')
        
        self._time = self._results[:,0]
        f.close()
 
    def __del__(self):
        warnings.warn("Extractor of file %s was deleted." % self._trajfilename)
    
    def getBus(self, busname):
        """Returns an object that allows to extract or plot bus related variables.
        
        :param str busname: the name of the bus

        :Example:

        >>> import pyramses
        >>> case = pyramses.cfg("case.rcfg") # load case from a configuration file
        >>> ram = pyramses.sim()
        >>> ram.execSim(case) # run the simulation
        >>> ext = pyramses.extractor(case.getTrj())
        >>> ext.getBus('1041').mag.plot() # will plot the timeseries simulated for the voltage magnitude on bus '1041' 
        
        .. note:: Available data are
           obsnames = ['mag','pha']
           obsdesc = ['Voltage magnitude (pu)','Voltage phase angle (deg)']
        """
        try:
            i=self._busname.index(busname) + 1 # +1 is to go to Fortran notation
            return self._getBusClass(self._time, self._results, 2*(i-1), busname)
        except ValueError:
            warnings.warn('Bus %s not found' % (busname))
    class _getBusClass(object):
        def _getElem(self, j, msg):
            tmp = self._shift + j
            return cur(self._time, self._results[:,tmp], msg)
        def __init__(self, time, results, shift, busname):
            self._time = time
            self._results = results
            self._shift = shift
            
            j=0
            self._obsnames = ['mag','pha']
            self._obsdesc = ['Voltage magnitude (pu)','Voltage phase angle (deg)']
            self.obsdict = dict(zip(self._obsnames, self._obsdesc))
            for name,msg in zip(self._obsnames, self._obsdesc):
                j=j+1
                setattr(self, name, self._getElem(j,busname+': '+msg))
                
    def getShunt(self, shuname):
        """Returns an object that allows to extract or plot shunt related variables.
        
        :param str shuname: the name of the shunt

        :Example:

        >>> import pyramses
        >>> case = pyramses.cfg("case.rcfg") # load case from a configuration file
        >>> ram = pyramses.sim()
        >>> ram.execSim(case) # run the simulation
        >>> ext = pyramses.extractor(case.getTrj())
        >>> ext.getShunt('sh1').Q.plot() # will plot the timeseries simulated for the reactive power of shunt 'sh1' 
        
        .. note:: Available data are
           obsnames = ['Q']
           obsdesc = ['Reactive power produced (Mvar)']
        
        """
        try:
            i=self._shuname.index(shuname) + 1 # +1 is to go to Fortran notation
            return self._getShuClass(self._time, self._results, 2*self._busnum + i-1, shuname)
        except ValueError:
            warnings.warn('Shunt %s not found' % (shuname))
            
    class _getShuClass(object):
        def _getElem(self, j, msg):
            tmp = self._shift + j
            return cur(self._time, self._results[:,tmp], msg)
        def __init__(self, time, results, shift, shuname):
            self._time = time
            self._results = results
            self._shift = shift     
            j=0
            self._obsnames = ['Q']
            self._obsdesc = ['Reactive power produced (Mvar)']
            self.obsdict = dict(zip(self._obsnames, self._obsdesc))
            for name,msg in zip(self._obsnames, self._obsdesc):
                j=j+1
                setattr(self, name, self._getElem(j,shuname+': '+msg))
        
    def getLoad(self, ldname):
        """Returns an object that allows to extract or plot load related variables.
        
        :param str ldname: the name of the load

        :Example:

        >>> import pyramses
        >>> case = pyramses.cfg("case.rcfg") # load case from a configuration file
        >>> ram = pyramses.sim()
        >>> ram.execSim(case) # run the simulation
        >>> ext = pyramses.extractor(case.getTrj())
        >>> ext.getLoad('L_1').P.plot() # will plot the timeseries simulated for the active power of 'L_1' 
        
        .. note:: Available data are
           obsnames = ['P','Q']
           obsdesc = ['Active power consumed (MW)','Reactive power consumed (Mvar)']
        """
        try:
            i=self._ldname.index(ldname) + 1 # +1 is to go to Fortran notation
            return self._getLdClass(self._time, self._results, 2*self._busnum+self._shunum +2*(i-1), ldname)
        except ValueError:
            warnings.warn('Load %s not found' % (ldname))
            
    class _getLdClass(object):
        def _getElem(self, j, msg):
            tmp = self._shift + j
            return cur(self._time, self._results[:,tmp], msg)
        def __init__(self, time, results, shift, ldname):
            self._time = time
            self._results = results
            self._shift = shift
            j=0
            self._obsnames = ['P','Q']
            self._obsdesc = ['Active power consumed (MW)','Reactive power consumed (Mvar)']
            self.obsdict = dict(zip(self._obsnames, self._obsdesc))
            for name,msg in zip(self._obsnames, self._obsdesc):
                j=j+1
                setattr(self, name, self._getElem(j,ldname+': '+msg))
  
    def getBranch(self, braname):
        """Returns an object that allows to extract or plot branch related variables.
        
        :param str braname: the name of the branch

        :Example:

        >>> import pyramses
        >>> case = pyramses.cfg("case.rcfg") # load case from a configuration file
        >>> ram = pyramses.sim()
        >>> ram.execSim(case) # run the simulation
        >>> ext = pyramses.extractor(case.getTrj())
        >>> ext.getBranch('1041-4041').PF.plot() # will plot the timeseries simulated for the active power of line'1041-4041' 
        
        .. note:: Available data are
           obsnames = ['PF','QF','PT','QT','RM','RA']
           obsdesc = ['P (MW) entering at FROM end', 'Q (Mvar) entering at FROM end',
                    'P (MW) entering at TO end', 'Q (Mvar) entering at TO end',
                    'magnitude of transformer ratio','phase angle of transformer ratio (deg)']
        
        """
        try:
            i=self._braname.index(braname) + 1 # +1 is to go to Fortran notation
            return self._getBraClass(self._time, self._results, 2*self._busnum+
                                 self._shunum+2*self._ldnum+6*(i-1), braname)
        except ValueError:
            warnings.warn('Branch %s not found' % (ldnabranameme))
            
    class _getBraClass(object):
        def _getElem(self, j, msg):
            tmp = self._shift + j
            return cur(self._time, self._results[:,tmp], msg)
        def __init__(self, time, results, shift, braname):
            self._time = time
            self._results = results
            self._shift = shift
            j=0
            self._obsnames = ['PF','QF','PT','QT','RM','RA']
            self._obsdesc = ['P (MW) entering at FROM end', 'Q (Mvar) entering at FROM end',
                    'P (MW) entering at TO end', 'Q (Mvar) entering at TO end',
                    'magnitude of transformer ratio','phase angle of transformer ratio (deg)']
            self.obsdict = dict(zip(self._obsnames, self._obsdesc))
            for name,msg in zip(self._obsnames, self._obsdesc):
                j=j+1
                setattr(self, name, self._getElem(j,braname+': '+msg))
        
    def getSync(self, syncname):
        """Returns an object that allows to extract or plot synchronous machine related variables.
        
        :param str syncname: the name of the sync machine

        :Example:

        >>> import pyramses
        >>> case = pyramses.cfg("case.rcfg") # load case from a configuration file
        >>> ram = pyramses.sim()
        >>> ram.execSim(case) # run the simulation
        >>> ext = pyramses.extractor(case.getTrj())
        >>> ext.getSync('g1').P.plot() # will plot the timeseries simulated for the active power of 'g1' 
        
        .. note:: 

           Available data are
           * obsnames = ['P','Q','A','S','FW','DD','QD','QW','FC','FV','T','ET','SC']
           * obsdesc = ['active power produced (MW)',
                    'reactive power produced (Mvar)',
                    'rotor angle wrt COI (deg)',
                    'rotor speed (pu)',
                    'flux in field winding (pu mach. base)     ',
                    'flux in d1 damper (pu mach. base)         ',
                    'flux in q1 damper (pu mach. base)         ',
                    'flux in q2 winding (pu mach. base)        ',
                    'field current (pu)                        ',
                    'field voltage (pu)                        ',
                    'mechanical torque (pu)                    ',
                    'electromagnetic torque (pu mach. base)    ',
                    'speed of COI reference (pu)               ']
        """
        try:
            i=self._syncname.index(syncname) + 1 # +1 is to go to Fortran notation
            return self._getSyncClass(self._time, self._results, 
                                  2*self._busnum+self._shunum+2*self._ldnum+6*self._branum+
                                  13*(i-1)+self._adexc[i-1]-1+self._adtor[i-1]-1, syncname)
        except ValueError:
            warnings.warn('Sync machine %s not found' % (syncname))
        
    class _getSyncClass(object):
        def _getElem(self, j, msg):
            tmp = self._shift + j
            return cur(self._time, self._results[:,tmp], msg)
        
        def __init__(self, time, results, shift, syncname):
            self._time = time
            self._results = results
            self._shift = shift
            j=0
            self._obsnames = ['P','Q','A','S','FW','DD','QD','QW','FC','FV','T','ET','SC']
            self._obsdesc = ['active power produced (MW)',
                    'reactive power produced (Mvar)',
                    'rotor angle wrt COI (deg)',
                    'rotor speed (pu)',
                    'flux in field winding (pu mach. base)     ',
                    'flux in d1 damper (pu mach. base)         ',
                    'flux in q1 damper (pu mach. base)         ',
                    'flux in q2 winding (pu mach. base)        ',
                    'field current (pu)                        ',
                    'field voltage (pu)                        ',
                    'mechanical torque (pu)                    ',
                    'electromagnetic torque (pu mach. base)    ',
                    'speed of COI reference (pu)               ']
            self.obsdict = dict(zip(self._obsnames, self._obsdesc))
            for name,msg in zip(self._obsnames, self._obsdesc):
                j=j+1
                setattr(self, name, self._getElem(j,syncname+': '+msg))
        
    def getExc(self, syncname):
        """Returns an object that allows to extract or plot exciter related variables.
        
        :param str syncname: the name of the generator that we want to check the exciter

        :Example:

        >>> import pyramses
        >>> case = pyramses.cfg("case.rcfg") # load case from a configuration file
        >>> ram = pyramses.sim()
        >>> ram.execSim(case) # run the simulation
        >>> ext = pyramses.extractor(case.getTrj())
        >>> ext.getExc('g1').vf.plot() # will plot the timeseries simulated for the field voltage of 'g1' 
        """
        try:
            i=self._syncname.index(syncname) + 1 # +1 is to go to Fortran notation
            return self._getExcClass(self._time, self._results, 
                                  2*self._busnum+self._shunum+2*self._ldnum+6*self._branum+
                                  13*(i-1)+self._adexc[i-1]-1+self._adtor[i-1]-1 + 13, self._excobsname[i-1], syncname)
        except ValueError:
            warnings.warn('Sync machine %s not found' % (syncname))
        
    class _getExcClass(object):
        
        def _getElem(self, j, msg):
            tmp = self._shift + j
            return cur(self._time, self._results[:,tmp], msg)
        
        def __init__(self, time, results, shift, excobsname, syncname):
            self._time = time
            self._results = results
            self._shift = shift
            self.obsdict = dict(zip(excobsname, ['User model, refer to the manual']*len(excobsname)))
            j=0
            for name in excobsname:
                j=j+1
                setattr(self, name, self._getElem(j,syncname+': '+name))
    
    def getTor(self, syncname):
        """Returns an object that allows to extract or plot governor related variables.
        
        :param str syncname: the name of the generator that we want to check the governor

        :Example:

        >>> import pyramses
        >>> case = pyramses.cfg("case.rcfg") # load case from a configuration file
        >>> ram.execSim(case) # run the simulation
        >>> ext = pyramses.extractor(case.getTrj())
        >>> ext.getTor('g1').Tm.plot() # will plot the timeseries simulated for the torque of 'g1' 
        """
        try:
            i=self._syncname.index(syncname) + 1 # +1 is to go to Fortran notation
            return self._getTorClass(self._time, self._results, 
                                  2*self._busnum+self._shunum+2*self._ldnum+6*self._branum+
                                  13*(i-1)+self._adexc[i]-1+self._adtor[i-1]-1 + 13, self._torobsname[i-1], syncname)
        except ValueError:
            warnings.warn('Sync machine %s not found' % (syncname))
        
    class _getTorClass(object):
        
        def _getElem(self, j, msg):
            tmp = self._shift + j
            return cur(self._time, self._results[:,tmp], msg)
        
        def __init__(self, time, results, shift, torobsname, syncname):
            self._time = time
            self._results = results
            self._shift = shift
            self.obsdict = dict(zip(torobsname, ['User model, refer to the manual']*len(torobsname)))
            j=0
            for name in torobsname:
                j=j+1
                setattr(self, name, self._getElem(j,syncname+': '+name))
                
    def getInj(self, injname):
        """Returns an object that allows to extract or plot injector related variables.
        
        :param str injname: the name of the injector

        :Example:

        >>> import pyramses
        >>> case = pyramses.cfg("case.rcfg") # load case from a configuration file
        >>> ram = pyramses.sim()
        >>> ram.execSim(case) # run the simulation
        >>> ext = pyramses.extractor(case.getTrj())
        >>> ext.getInj('pv1').P.plot() # will plot the timeseries simulated for the active power of 'pv1' 
        """
        try:
            i=self._injname.index(injname) + 1 # +1 is to go to Fortran notation
            return self._getInjClass(self._time, self._results, 
                                 2*self._busnum+self._shunum+2*self._ldnum+6*self._branum+
                                 13*(self._syncnum)+self._adexc[self._syncnum]-1+
                                 self._adtor[self._syncnum]-1+self._adinj[i-1]-1, 
                                 self._injobsname[i-1], injname)
        except ValueError:
            warnings.warn('Injector %s not found' % (injname))
        
    class _getInjClass(object):
        
        def _getElem(self, j, msg):
            tmp = self._shift + j
            return cur(self._time, self._results[:,tmp], msg)
        
        def __init__(self, time, results, shift, injobsname, injname):
            self._time = time
            self._results = results
            self._shift = shift
            self.obsdict = dict(zip(injobsname, ['User model, refer to the manual']*len(injobsname)))
            j=0
            for name in injobsname:
                j=j+1
                setattr(self, name, self._getElem(j,injname+': '+name))
    
    def getTwop(self, twopname):
        """Returns an object that allows to extract or plot twoport related variables.
        
        :param str twopname: the name of the twoport

        :Example:

        >>> import pyramses
        >>> case = pyramses.cfg("case.rcfg") # load case from a configuration file
        >>> ram = pyramses.sim()
        >>> ram.execSim(case) # run the simulation
        >>> ext = pyramses.extractor(case.getTrj())
        >>> ext.getTwop('lcc1').P1.plot() # will plot the timeseries simulated for the power of 'lcc1' 
        """
        try:
            i=self._twopname.index(twopname) + 1 # +1 is to go to Fortran notation
            return self._getTwopClass(self._time, self._results, 
                                 2*self._busnum+self._shunum+2*self._ldnum+6*self._branum+
                                 13*(self._syncnum)+self._adexc[self._syncnum]-1+
                                 self._adtor[self._syncnum]-1+self._adinj[self._injnum]-1+
                                 self._adtwop[i-1]-1, 
                                 self._twopobsname[i-1], twopname)
        except ValueError:
            warnings.warn('Twoport %s not found' % (twopname))
        
    class _getTwopClass(object):
        
        def _getElem(self, j, msg):
            tmp = self._shift + j
            return cur(self._time, self._results[:,tmp], msg)
        
        def __init__(self, time, results, shift, twopobsname, twopname):
            self._time = time
            self._results = results
            self._shift = shift
            self.obsdict = dict(zip(twopobsname, ['User model, refer to the manual']*len(twopobsname)))
            j=0
            for name in twopobsname:
                j=j+1
                setattr(self, name, self._getElem(j,twopname+': '+name))
    
    def getDctl(self, dctlname):
        """Returns an object that allows to extract or plot dctl related variables.
        
        :param str dctlname: the name of the dctl

        :Example:

        >>> import pyramses
        >>> case = pyramses.cfg("case.rcfg") # load case from a configuration file
        >>> ram = pyramses.sim()
        >>> ram.execSim(case) # run the simulation
        >>> ext = pyramses.extractor(case.getTrj())
        >>> ext.getDctl('agc').g5.plot() # will plot the timeseries simulated for the power of 'g5' 
        """
        try:
            i=self._dctlname.index(dctlname) + 1 # +1 is to go to Fortran notation
            return self._getDCTLClass(self._time, self._results, 
                                 2*self._busnum+self._shunum+2*self._ldnum+6*self._branum+
                                 13*(self._syncnum)+self._adexc[self._syncnum]-1+
                                 self._adtor[self._syncnum]-1+self._adinj[self._injnum]-1+
                                 self._adtwop[self._twopnum]-1+self._addctl[i-1]-1, 
                                 self._dctlobsname[i-1],dctlname)
        except ValueError:
            warnings.warn('DCTL %s not found' % (dctlname))
        
    class _getDCTLClass(object):
        
        def _getElem(self, j, msg):
            tmp = self._shift + j
            return cur(self._time, self._results[:,tmp], msg)
        
        def __init__(self, time, results, shift, dctlobsname, dctlname):
            self._time = time
            self._results = results
            self._shift = shift
            self.obsdict = dict(zip(dctlobsname, ['User model, refer to the manual']*len(dctlobsname)))
            j=0
            for name in dctlobsname:
                j=j+1
                setattr(self, name, self._getElem(j,dctlname+': '+name))


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Python library for RAMSES dynamic simulator."""

__name__ = "pyramses"
__version__ = '0.0.27'
__author__ = "Petros Aristidou"
__copyright__ = "Petros Aristidou"
__license__ = "Petros Aristidou"
__maintainer__ = "Petros Aristidou"
__email__ = "apetros@pm.me"
__url__ = "https://pyramses.sps-lab.org"
__status__ = "3 - Alpha"

import sys
from warnings import warn

from .cases import cfg
from .globals import __runTimeObs__, __which
from .simulator import sim
from .extractor import extractor, curplot, cur

if sys.platform in ('win32', 'cygwin'):
    checkGnuplot = __which('gnuplot.exe')
else:
    checkGnuplot = __which('gnuplot')
if checkGnuplot is None:
    warn("RAMSES: Gnuplot executable could not be found in the system path, so the runtime observables are disabled.")
    __runTimeObs__ = False
else:
    __runTimeObs__ = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Python module for RAMSES dynamic simulator. It provides an interface through Python to
perform dynamic simulations and extract data from RAMSES."""

import ctypes
import _ctypes
import os
import sys
import warnings
import datetime
import numpy as np
from scipy.sparse import coo_matrix

from .cases import cfg
from .globals import RAMSESError, CustomWarning, __libdir__, wrapToList

class sim(object):
    """Simulation class."""

    warnings.showwarning = CustomWarning

    ramsesCount = 0  # Provides a sum of all instances of Ramses running
    os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"  # This allows to run even if 2 versions of MKL are present on the computer

    # print(__libdir__)
    # print(os.path.join(__libdir__,"ramses.dll"))

    def __init__(self, custLibDir = None):
        """Initialize the DLL libraries."""
        if custLibDir is None:
            ramLibDir = __libdir__
        else:
            try:
                if os.path.exists(custLibDir) and os.path.isdir(custLibDir):
                    ramLibDir = custLibDir
                    warnings.warn("Overwriting the internal DLL with the one at %s. The DLL in that location will be locked. To unlock it, you need to del() this instance of pyramses.sim() or restart the kernel." % ramLibDir)
                else:
                    raise RAMSESError('RAMSES: The path %s does not exist or it is not a directory.' % (custLibDir))
            except OSError as e:
                raise RAMSESError('RAMSES: The path %s gave an error: ' % (custLibDir), e)
            
        try:
            # print(sys.platform)
            if sys.platform in ('win32', 'cygwin'):
                self._ramseslib = ctypes.CDLL(os.path.join(ramLibDir, "ramses.dll"))
            else:
                # if (not find_library('iomp5')):
                # self._omplib = ctypes.CDLL(os.path.join(__libdir__, "libiomp5.so"), mode=ctypes.RTLD_GLOBAL)
                # self._mkllib = ctypes.CDLL(os.path.join(__libdir__, "libmkl.so"), mode=ctypes.RTLD_GLOBAL)
                self._ramseslib = ctypes.CDLL(os.path.join(ramLibDir, "ramses.so"))
        except OSError as e:
            raise ImportError('RAMSES: ', e)

        sim.ramsesCount += 1
        self._ramsesNum = sim.ramsesCount  # This is the current instance of Ramses
        self._setcalls()

    def __del__(self):
        warnings.warn("Simulator with number %i was deleted." % self._ramsesNum)
        if sys.platform in ('win32', 'cygwin'):
            _ctypes.FreeLibrary(self._ramseslib._handle)
        else:
            _ctypes.dlclose(self._ramseslib._handle)
        sim.ramsesCount -= 1

    def _c_func_wrapper(self, cdecl_text):
        """Read the ramses.h file and sets the necessary call types for Python."""

        def move_pointer_and_strip(type_def, name):
            if '*' in name:
                type_def += ' ' + name[:name.rindex('*') + 1]
                name = name.rsplit('*', 1)[1]
            return type_def.strip(), name.strip()

        def type_lookup(type_def):
            '''Supported C variable types and their ctypes equivalents.'''
            types = {
                'void': None,
                'char *': ctypes.c_char_p,
                'int': ctypes.c_int,
                'int *': ctypes.POINTER(ctypes.c_int),
                'void *': ctypes.c_void_p,
                'size_t': ctypes.c_size_t,
                'size_t *': ctypes.POINTER(ctypes.c_size_t),
                'double': ctypes.c_double,
                'double *': ctypes.POINTER(ctypes.c_double)
            }
            type_def_without_const = type_def.replace('const ', '')
            if type_def_without_const in types:
                return types[type_def_without_const]
            elif (type_def_without_const.endswith('*') and type_def_without_const[:-1] in types):
                return ctypes.POINTER(types[type_def_without_const[:-1]])
            else:
                raise KeyError(type_def)

        a, b = [i.strip() for i in cdecl_text.split('(', 1)]
        params, _ = b.rsplit(')', 1)
        rtn_type, name = move_pointer_and_strip(*a.rsplit(' ', 1))
        param_spec = []
        for param in params.split(','):
            if param != 'void':
                param_spec.append(move_pointer_and_strip(*param.rsplit(' ', 1)))

        try:
            func = getattr(self._ramseslib, name)  # get the function from the dll
            setattr(func, 'restype', type_lookup(rtn_type))  # set the return type
            setattr(func, 'argtypes', [type_lookup(type_def) for type_def, _ in param_spec])  # set the argument types
        except AttributeError as e:
            #raise AttributeError(
            #    'RAMSES: Function %s is listed in ramses.h but cannot be found in the library.' % (name), e)
            warnings.warn('RAMSES: Function %s is listed in ramses.h but cannot be found in the library.' % (name))
            print(e)

    def _setcalls(self):
        """Define the argument types and returning types for the routines in RAMSES"""
        try:
            with open(os.path.join(__libdir__, "ramses.h"), 'r') as f:
                _C_HEADER = f.read()
                for cdecl_text in _C_HEADER.splitlines():
                    if cdecl_text.strip():
                        if not cdecl_text.startswith("//"):
                            self._c_func_wrapper(cdecl_text)
        except IOError as e:
            raise IOError("RAMSES: Cannot open ramses.h files", e)

    def getLastErr(self):
        """Return the last error message issued by RAMSES.

        :rtype: str

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case, 0) # start simulation paused
        >>> ram.getLastErr()
        'This is an error msg'

        """

        errMsg = ctypes.create_string_buffer(1024)

        try:
            retval = self._ramseslib.get_last_err_log(errMsg)
        except KeyError:
            retval = 1

        if (retval != 0):
            raise RAMSESError('RAMSES: Function getLastErr failed.')

        return errMsg.value.decode()

    def getJac(self):
        """Gets the Jacobian matrix written in files

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case, 0) # start simulation paused
        >>> ram.getJac()

        """
        try:
            retval = self._ramseslib.get_Jac()
        except KeyError:
            retval = 1

        if (retval != 0) and (retval != 112):
            raise RAMSESError('RAMSES: Function get_Jac failed.')
        
        Nmat = 0
        try:
            f = open('py_eqs.dat','r')
            lines = f.readlines()
            f.close()
            rowl = []
            coll = []
            datal = []
            for x in lines:
                xsplit = x.split()
                if int(xsplit[0]) > Nmat:
                    Nmat = int(xsplit[0])
                if int(xsplit[5]) > 0:
                    rowl.append(int(xsplit[0])-1)
                    coll.append(int(xsplit[5])-1)
                    datal.append(1)
            row = np.array( rowl, dtype=int )
            col = np.array( coll, dtype=int )
            data = np.array( datal, dtype=float )
            E = coo_matrix((data,(row,col)), shape=(Nmat,Nmat)).tocsc()
        except:
            raise RAMSESError('RAMSES: Function get_Jac failed while reading E.')
            
        try:
            f = open('py_val.dat','r')
            lines = f.readlines()
            f.close()
            row = np.array( [ int(x.split()[0])-1 for x in lines ], dtype=int )
            col = np.array( [ int(x.split()[1])-1 for x in lines ], dtype=int )
            data = np.array( [ x.split()[2] for x in lines ], dtype=float )
            A = coo_matrix((data,(row,col)), shape=(Nmat,Nmat)).tocsc()
        except:
            raise RAMSESError('RAMSES: Function get_Jac failed while reading A.')

        return A, E

    def getCompName(self, comp_type, num):
        """Return the name of the num:superscript:`th` component of type comp_type.

        :param str comp_type: the component type (BUS, SYNC, INJ, DCTL, BRANCH, TWOP, SHUNT, LOAD)
        :param int num: The number of the component to be fetched
        :returns: The component name
        :rtype: srt

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case, 0) # start simulation paused
        >>> ram.getCompName("BUS",1)
        'B1'

        """
        get_name_func = {
            'BUS': self._ramseslib.get_bus_name,
            'SYNC': self._ramseslib.get_sync_name,
            'INJ': self._ramseslib.get_inj_name,
            'DCTL': self._ramseslib.get_dctl_name,
            'BRANCH': self._ramseslib.get_branch_name,
            'TWOP': self._ramseslib.get_twop_name,
            'SHUNT': self._ramseslib.get_shunt_name,
            'LOAD': self._ramseslib.get_load_name
        }

        name = ctypes.create_string_buffer(21)

        try:
            retval = get_name_func[comp_type](num, name)
        except KeyError:
            retval = 1

        if (retval != 0):
            raise RAMSESError(
                'RAMSES: Function getCompName(%s,%i) failed. Does this component exist?' % (comp_type, num))

        return name.value.decode()

    def getAllCompNames(self, comp_type):
        """Get list of all components of type comp_type

        :param str comp_type: the component type (BUS, SYNC, INJ, DCTL, BRANCH, TWOP, SHUNT, LOAD)
        :returns: list of component names
        :rtype: list of str

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case, 0) # start simulation paused
        >>> ram.getAllCompNames("BUS")
        ['B1',
        'B2',
        'B3']

        """
        get_nb_func = {
            'BUS': self._ramseslib.get_nbbus,
            'SYNC': self._ramseslib.get_nbsync,
            'INJ': self._ramseslib.get_nbinj,
            'DCTL': self._ramseslib.get_nbdctl,
            'BRANCH': self._ramseslib.get_nbbra,
            'TWOP': self._ramseslib.get_nbtwop,
            'SHUNT': self._ramseslib.get_nbshunt,
            'LOAD': self._ramseslib.get_nbload
        }

        try:
            Nb = get_nb_func[comp_type]()
        except KeyError:
            raise RAMSESError(
                'RAMSES: Function getAllCompName(%s) failed. Does this component type exists?' % (comp_type))

        names = []
        if Nb > 0:
            for comp in range(1, Nb + 1):
                names.append(self.getCompName(comp_type, comp))
        return names

    def execSim(self, cmd, pause=None):
        """Run a simulation.

        :param cmd: provides the case description
        :type cmd: type(:class:`.cfg`)
        :param t_pause: pause time (optional). If not given, the simulation will run until the end as described in the dst file.
        :type t_pause: float

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case) # start simulation

        .. note:: If you have an existing command file, you can pass it to the simulator as pyramses.cfg("cmd.txt")
        """
        if not isinstance(cmd, cfg):
            raise TypeError('RAMSES: Function execSim because the command file is not of type pyramses.cfg()')
        if pause is not None:
            self.pauseSim(pause)
        cmdfilename = cmd.writeCmdFile()
        if not cmd._out:
            outfilename = os.path.join(os.getcwd(),'output '+datetime.datetime.now().strftime("%d.%m.%Y-%H.%M.%S")+'.trace')
        else:
            outfilename = cmd._out[0]
        retval = self._ramseslib.ramses(cmdfilename.encode('utf-8'), outfilename.encode('utf-8'))
        if (retval != 0) and (retval != 112):
            raise RAMSESError('RAMSES: Function execSim() failed with the flag %i. Last message was: %s' % (retval, self.getLastErr()))
        return 0

    def contSim(self, pause=None):
        """Continue the simulation until the given time.

        The pause argument is optional and it gives the time until the simulation will stop."""
        if pause is not None:
            self.pauseSim(pause)
        retval = self._ramseslib.continue_simul()
        if (retval != 0) and (retval != 112):
            raise RAMSESError('RAMSES: Function contSim() failed with the flag %i. Last message was: %s' % (retval, self.getLastErr()))
        return 0

    def getBusVolt(self, busNames):
        """Return the voltage magnitude of a list of buses

        :param busNames: the names of buses
        :type busNames: list of str
        :returns: list of bus voltage magnitudes
        :rtype: list of floats

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case, 10.0) # simulate until 10 seconds and pause
        >>> buses = ['g1','g2','g3']
        >>> ram.getBusVolt(buses)
        [1.0736851673414456,
         1.0615442362180327,
         1.064702686997689]
        """
        volts = []
        bus_volt = ctypes.c_double()
        for bus in busNames:
            retval = self._ramseslib.get_volt_mag(bus.encode('utf-8'), bus_volt)
            if (retval != 0):
                raise RAMSESError(
                    'RAMSES: Function getBusVolt(%s) failed with the flag %i.  Does the bus exist? Last message was: %s' % (bus, retval, self.getLastErr()))
            volts.append(bus_volt.value)
        return volts

    def getBranchPow(self, branchName):
        """Return the active and reactive powers of a list of branches

        :param branchName: the names of branches
        :type branchName: list of str
        :returns: list of branch powers. These are active and reactive power at the origin and extremity respectively (p_orig, q_orig, p_extr, q_extr)
        :rtype: list of floats

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case, 10.0) # simulate until 10 seconds and pause
        >>> branchName = ['1011-1013','1012-1014','1021-1022']
        >>> ram.getBranchPow(branchName)
        """
        pows = []
        p_orig = ctypes.c_double()
        q_orig = ctypes.c_double()
        p_extr = ctypes.c_double()
        q_extr = ctypes.c_double()
        for branch in branchName:
            retval = self._ramseslib.get_line_pow(branch.encode('utf-8'), p_orig, q_orig, p_extr, q_extr)
            if (retval != 0):
                raise RAMSESError(
                    'RAMSES: Function get_line_pow(%s) failed with the flag %i.  Does the branch exist? Last message was: %s' % (branch, retval, self.getLastErr()))
            thisBranch = [p_orig.value, q_orig.value, p_extr.value, q_extr.value]
            pows.append(thisBranch)
        return pows

    def getBranchCur(self, branchName):
        """Return the currents of a list of branches

        :param branchName: the names of branches
        :type branchName: list of str
        :returns: list of branch currents. These are x-y components at the origin and extremity respectively (ix_orig, iy_orig, ix_extr, iy_extr)
        :rtype: list of floats

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case, 10.0) # simulate until 10 seconds and pause
        >>> branchName = ['1011-1013','1012-1014','1021-1022']
        >>> ram.getBranchCur(branchName)

        """
        curs = []
        ix_orig = ctypes.c_double()
        iy_orig = ctypes.c_double()
        ix_extr = ctypes.c_double()
        iy_extr = ctypes.c_double()
        for branch in branchName:
            retval = self._ramseslib.get_line_cur(branch.encode('utf-8'), ix_orig, iy_orig, ix_extr, iy_extr)
            if (retval != 0):
                raise RAMSESError(
                    'RAMSES: Function get_line_cur(%s) failed with the flag %i.  Does the branch exist? Last message was: %s' % (branch, retval, self.getLastErr()))
            thisBranch = [ix_orig.value, iy_orig.value, ix_extr.value, iy_extr.value]
            curs.append(thisBranch)
        return curs        

    def getBusPha(self, busNames):
        """Return the voltage phase of a list of buses

        :param busNames: the names of buses
        :type busNames: list of str
        :returns: list of bus voltage phase
        :rtype: list of floats

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case, 10.0) # simulate until 10 seconds and pause
        >>> buses = ['g1','g2','g3']
        >>> ram.getBusPha(buses)
        [0.0000000000000000,
         10.0615442362180327,
         11.064702686997689]

        """
        pha = []
        bus_pha = ctypes.c_double()
        for bus in busNames:
            retval = self._ramseslib.get_volt_pha(bus.encode('utf-8'), bus_pha)
            if (retval != 0):
                raise RAMSESError(
                    'RAMSES: Function getBusPha(%s) failed with the flag %i.  Does the bus exist? Last message was: %s' % (bus, retval, self.getLastErr()))
            pha.append(bus_pha.value)
        return pha

    def endSim(self):
        """End the simulation"""

        self._ramseslib.set_end_simul()
        return self.contSim(self.getInfTime())
    
    def getEndSim(self):
        """Check if the simulation has ended
        
        :returns: 0 -> simulation is still working, 1 -> simulation has ended
        :rtype: integer
        
        """

        return self._ramseslib.get_end_simul()

    def getSimTime(self):
        """Get the current simulation time

        :returns: current simulation time in RAMSES.
        :rtype: float

        """
        return self._ramseslib.get_sim_time()

    def getInfTime(self):
        """Get the maximum representable double from the simulator ()

        :returns: maximum representable double
        :rtype: float

        """
        return self._ramseslib.get_huge_double()

    def initObserv(self,traj_filenm):
        """Initialize observable selection (structure and trajectory file)

        :param traj_filenm: the file to save the trajectory at
        :type traj_filenm: str

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt") # command file without any observables
        >>> ram.execSim(case, 0.0) # start
        >>> traj_filenm = 'obs.trj'
        >>> ram.initObserv(traj_filenm)
        """

        return self._ramseslib.initObserv(traj_filenm.encode('utf-8'))

    def addObserv(self,string):
        """Add an element to be observed

        :param string: the element with proper format
        :type string: str

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt") # command file without any observables
        >>> ram.execSim(case, 0.0) # start
        >>> traj_filenm = 'obs.trj'
        >>> ram.initObserv(traj_filenm)
        >>> string = 'BUS *' # monitor all buses
        >>> ram.addObserv(string)
        """

        return self._ramseslib.addObserv(string.encode('utf-8'))

    def finalObserv(self):
        """Finalize observable selection, allocate buffer, and write header of trajectory file

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt") # command file without any observables
        >>> ram.execSim(case, 0.0) # start
        >>> traj_filenm = 'obs.trj'
        >>> ram.initObserv(traj_filenm)
        >>> string = 'BUS *' # monitor all buses
        >>> ram.addObserv(string)
        >>> ram.finalObserv()
        """

        return self._ramseslib.finalObserv()

    def getPrm(self, comp_type, comp_name, prm_name):
        """Get the value of a named parameter

        :param comp_type: the types of components (EXC, TOR, INJ, DCTL, TWOP)
        :type comp_type: list of str
        :param comp_name: the names of components
        :type comp_name: list of str
        :param prm_name: the names of parameters
        :type prm_name: list of str

        :returns: list of parameter values
        :rtype: list of floats

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case, 10.0) # simulate until 10 seconds and pause
        >>> comp_type = ['EXC','EXC','EXC']
        >>> comp_name = ['g1','g2','g3']
        >>> prm_name = ['V0','V0','V0']
        >>> ram.getPrm(comp_type,comp_name, prm_name)
        [1.0736851673414456,
         1.0615442362180327,
         1.064702686997689]

        """

        comp_type = wrapToList(comp_type)
        comp_name = wrapToList(comp_name)
        prm_name = wrapToList(prm_name)

        if not (len(comp_type) == len(comp_name) == len(prm_name)):
            raise ValueError('RAMSES: Function getPrm failed because the lists are not equal!')
        prm_values = []
        for a, b, c in zip(comp_type, comp_name, prm_name):
            prm_value = ctypes.c_double()
            retval = self._ramseslib.get_named_prm(a.encode('utf-8'), b.encode('utf-8'), c.encode('utf-8'),
                                                   prm_value)
            if retval != 0:
                raise RAMSESError(
                    'RAMSES: Function getPrm(%s,%s,%s) failed with the flag %i. Does the parameter or equipment exist? Last message was: %s'
                    % (a, b, c, retval, self.getLastErr()))
            prm_values.append(prm_value.value)
        if len(prm_values) == 1:
            return prm_values[0]
        else:
            return prm_values

    def defineSS(self, ssID, filter1, filter2, filter3):
        """Define a subsytem using three filters. The resulting list is an intersection of the filters.

        :param ssID: Number of the SS
        :type ssID: int
        :param filter1: Voltage levels to be included
        :type filter1: list of str or str
        :param filter2: Zones (which zone/zones will be included)
        :type filter2: list of str or str
        :param filter3: Bus names to be included
        :type filter3: list of str or str

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case, 0.0)
        >>> ram.defineSS(1, ['735'], [], []) # SS 1 with all buses at 735 kV, no zones, no list of buses

        .. note:: An empty filter means it is deactivated and discarded.

        """
        filter1 = wrapToList(filter1)
        filter2 = wrapToList(filter2)
        filter3 = wrapToList(filter3)

        strF1 = ' '.join(filter1)
        strF2 = ' '.join(filter2)
        strF3 = ' '.join(filter3)

        retval = self._ramseslib.define_SS(ctypes.c_int(ssID), strF1.encode('utf-8'), ctypes.c_int(len(filter1)),
                                           strF2.encode('utf-8'), ctypes.c_int(len(filter2)), strF3.encode('utf-8'),
                                           ctypes.c_int(len(filter3)))
        if retval != 0:
            raise RAMSESError('RAMSES: Function define_SS(%i,...) failed with the flag %i. Last message was: %s' % (ssID, retval, self.getLastErr()))

    def getSS(self, ssID):
        """Retrieve the buses of a subsytem.

        :param ssID: Number of the SS
        :type ssID: int

        :returns: list of buses
        :rtype: list of str

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case, 0.0)
        >>> ram.defineSS(1, ['735'], [], []) # SS 1 with all buses at 735 kV
        >>> ram.getSS(1) # get list of buses in SS 1

        """

        mxreclen = self._ramseslib.get_nbbus()
        string_buffer = ctypes.create_string_buffer(19 * mxreclen)

        retval = self._ramseslib.get_SS(ctypes.c_int(ssID), ctypes.c_int(mxreclen), string_buffer)
        if retval != 0:
            raise RAMSESError('RAMSES: Function getSS(%i) failed with the flag %i. Last message was: %s' % (ssID, retval, self.getLastErr()))

        return string_buffer.value.split()

    def getTrfoSS(self, ssID, location, in_service, rettype):
        """Retrieve transformer information of subsystem after applying some filters

        :param ssID: Number of the SS
        :type ssID: int
        :param location: 1 – both buses inside SS, 2 - tie transformers, 3 – 1 & 2
        :type location: int
        :param in_service: 1 – transformers in service, 2 - all transformers
        :type in_service: int
        :param rettype: Type of response (NAME, From, To, Status, Tap, Currentf, Currentt, Pf, Qf, Pt, Qt).
        :type rettype: str

        :returns: list of transformer names
        :rtype: list of str

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case, 0.0)
        >>> ram.defineSS(1, ['735'], [], []) # SS 1 with all buses at 735 kV
        >>> ram.getTrfoSS(1,3,2,'Status')

        .. note:: Tap is not implemented yet.

        .. todo:: Implement Taps after discussing with Lester.
        """

        if rettype not in ['NAME', 'From', 'To', 'Status', 'Tap', 'Currentf', 'Currentt', 'Pf', 'Qf', 'Pt', 'Qt']:
            raise RAMSESError('RAMSES: Function getTrfoSS(%i,%i,%i,%s): rettype is not valid!' % (
                ssID, location, in_service, rettype))
        elif location not in [1, 2, 3]:
            raise RAMSESError('RAMSES: Function getTrfoSS(%i,%i,%i,%s): location is not valid!' % (
                ssID, location, in_service, rettype))
        elif in_service not in [1, 2]:
            raise RAMSESError('RAMSES: Function getTrfoSS(%i,%i,%i,%s): in_service is not valid!' % (
                ssID, location, in_service, rettype))

        mxreclen = self._ramseslib.get_nbbra()
        string_buffer = ctypes.create_string_buffer(21 * mxreclen)
        dp_vec = (ctypes.c_double * mxreclen)()
        dp_int = (ctypes.c_int * mxreclen)()
        elem = ctypes.c_int()

        retval = self._ramseslib.get_transformer_in_SS(ctypes.c_int(ssID), ctypes.c_int(location),
                                                       ctypes.c_int(in_service), rettype.encode('utf-8'),
                                                       ctypes.c_int(mxreclen), string_buffer, dp_vec, dp_int, elem)
        if retval != 0:
            raise RAMSESError('RAMSES: Function getTrfoSS(%i,%i,%i,%s) failed with the flag %i.' % (
                ssID, location, in_service, rettype, retval))

        if elem == 0:
            return []
        elif rettype in ['Tap', 'Status']:
            return [dp_int[i] for i in range(0, elem.value)]
        elif rettype in ['Currentf', 'Currentt', 'Pf', 'Qf', 'Pt', 'Qt']:
            return [dp_vec[i] for i in range(0, elem.value)]
        else:
            return string_buffer.value.split()

    def getPrmNames(self, comp_type, comp_name):
        """Get the named parameters of a model

        :param comp_type: the types of components (EXC, TOR, INJ, DCTL, TWOP)
        :type comp_type: list of str or str
        :param comp_name: the names of component instances
        :type comp_name: list of str or str

        :returns: list of parameter names
        :rtype: list of lists of strings

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case, 0.0) # initialize and wait
        >>> comp_type = ['EXC','EXC','EXC']
        >>> comp_name = ['g1','g2','g3'] # name of synchronous machines
        >>> ram.getPrmNames(comp_type,comp_name)

        """
        comp_type = wrapToList(comp_type)
        comp_name = wrapToList(comp_name)

        if not (len(comp_type) == len(comp_name)):
            raise ValueError('RAMSES: Function getPrmNames failed because the lists are not equal!')
        prm_names = []
        mxprm = self._ramseslib.get_mxprm()  # get maximum number of parameters in any model
        for a, b in zip(comp_type, comp_name):
            string_buffer = ctypes.create_string_buffer(11 * mxprm)
            retval = self._ramseslib.get_comp_prm_names(a.encode('utf-8'), b.encode('utf-8'), ctypes.c_int(mxprm),
                                                        string_buffer)
            if retval != 0:
                raise RAMSESError(
                    'RAMSES: Function getPrm(%s,%s) failed with the flag %i. Does the parameter or equipment exist?'
                    % (a, b, retval))
            decodedstring = string_buffer.value.decode()
            prm_names.append(decodedstring.split())
        if len(prm_names) == 1:
            return prm_names[0]
        else:
            return prm_names

    def getObs(self, comp_type, comp_name, obs_name):
        """Get the value of a named observable. 

        :param comp_type: the types of components ('EXC','TOR','INJ','TWOP','DCTL','SYN')
        :type comp_type: list of str
        :param comp_name: the names of components
        :type comp_name: list of str
        :param obs_name: the names of observables
        :type obs_name: list of str

        :returns: list of observable values
        :rtype: list of floats

        .. note:: 
          
          For the synchronous generator ('SYN') the accepted obs_name values are 
          - 'P': Active power (MW)
          - 'Q': Reactive power (MVAr)
          -'Omega': Machine speed (pu)
          - 'S': Apparent power (MVA)
          - 'SNOM': Nominal apparent power (MVA)
          - 'PNOM': Nominal active power (MW)

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case, 10.0) # simulate until 10 seconds and pause
        >>> comp_type = ['INJ','EXC','TOR']
        >>> comp_name = ['L_11','g2','g3']
        >>> obs_name = ['P','vf','Pm']
        >>> ram.getObs(comp_type,comp_name, obs_name)
        [199.73704732259995,
        1.3372945282585218,
        0.816671075203357]

        """
        comp_type = wrapToList(comp_type)
        comp_name = wrapToList(comp_name)
        obs_name = wrapToList(obs_name)

        if not (len(comp_type) == len(comp_name) == len(obs_name)):
            raise ValueError('RAMSES: Function getObs failed because the lists are not equal!')
        obs_values = []
        for a, b, c in zip(comp_type, comp_name, obs_name):
            obs_value = ctypes.c_double()
            retval = self._ramseslib.get_named_obs(a.encode('utf-8'), b.encode('utf-8'), c.encode('utf-8'),
                                                   obs_value)
            if retval != 0:
                raise RAMSESError(
                    'RAMSES: Function getObs(%s,%s,%s) failed with the flag %i. Does the observable or equipment exist?'
                    % (a, b, c, retval))
            obs_values.append(obs_value.value)
        return obs_values

    def pauseSim(self, t_pause):
        """Pause the simulation at the t_pause time

        :param t_pause: pause time
        :type t_pause: float

        """
        return self._ramseslib.set_pause_time(t_pause)

    def addDisturb(self, t_dist, disturb):
        """Add a new disturbance at a specific time. Follows the same structure as the disturbances in the dst files.

        :param t_dist: time of the disturbance
        :type t_dist: float
        :param disturb: description of disturbance
        :type disturb: str

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case, 80.0) # simulate until 80 seconds and pause
        >>> ram.addDisturb(100.000, 'CHGPRM DCTL 1-1041  Vsetpt -0.05 0') # Decrease the setpoint of the DCTL by 0.015 pu, at t=100 s
        >>> ram.addDisturb(100.000, 'CHGPRM DCTL 2-1042  Vsetpt -0.05 0')
        >>> ram.addDisturb(100.000, 'CHGPRM DCTL 3-1043  Vsetpt -0.05 0')
        >>> ram.addDisturb(100.000, 'CHGPRM DCTL 4-1044  Vsetpt -0.05 0')
        >>> ram.addDisturb(100.000, 'CHGPRM DCTL 5-1045  Vsetpt -0.05 0')
        >>> ram.contSim(ram.getInfTime()) # continue the simulation
        """
        return self._ramseslib.add_disturb(t_dist, disturb.encode('utf-8'))

    def load_MDL(MDLName):
        """Load external DLL file with user defined models. Should be in current directory or absolute path.

        :param MDLName: path to file
        :type MDLName: str

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> ram = pyramses.load_MDL("MDLs.dll")
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case) # simulate
        """
        return self._ramseslib.c_load_MDL(MDLName.encode('utf-8'))

    def unload_MDL(MDLName):
        """Unload external DLL file with user defined models. Should be in current directory or absolute path.

        :param MDLName: path to file
        :type MDLName: str

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> ram = pyramses.load_MDL("MDLs.dll")
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case) # simulate
        >>> ram = pyramses.unload_MDL("MDLs.dll")
        """
        return self._ramseslib.c_unload_MDL(MDLName.encode('utf-8'))


    def get_MDL_no():
        """Unload external DLL file with user defined models. Should be in current directory or absolute path.

        :returns: list of observable values
        :rtype: list of floats

        :Example:

        >>> import pyramses
        >>> ram = pyramses.sim()
        >>> ram = pyramses.load_MDL("MDLs.dll")
        >>> case = pyramses.cfg("cmd.txt")
        >>> ram.execSim(case) # simulate
        >>> ram = pyramses.get_MDL_no()
        """
        return self._ramseslib.c_get_MDL_no()
        

    
#int get_MDL_names(int mxreclen, char *DLL_Names);


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Contains the class to describe a test-case scenario in pyramses."""

import os
import tempfile
import warnings

from .globals import RAMSESError, __runTimeObs__, CustomWarning, silentremove

warnings.showwarning = CustomWarning


class cfg(object):
    """Test case description class."""

    def __init__(self, cmd=None):
        self._out = []  # output file
        self._dataset = []  # data files
        self._dstset = []  # disturbance files
        self._runobs = []  # runtime observables
        self._init = []  # initialization file
        self._disc = []  # discrete trace file
        self._cont = []  # continuous trace file
        self._obs = []  # observables file
        self._trj = []  # trajectory file
        #self._cmdfile = tempfile.NamedTemporaryFile(prefix='pyramses_', suffix='.txt', dir=os.getcwd(),
        #                                            delete=False)  # text file to pass to Ramses
        #self._cmdfile.close()

        if cmd:
            try:
                typeread = 1  # 1=data, 2=init, 3=dst, 4=trj, 5=obs,6=cont, 7=disc, 8=runtime obs
                with open(cmd) as data:
                    content = data.read().splitlines()
                    for line in content:
                        # print line
                        token = line.split()
                        if typeread == 1:
                            if len(token) > 0:
                                self.addData(line)
                            else:
                                if not self._dataset:
                                    raise Exception('RAMSES: At least one datafile is necessary in the command file.')
                                    return
                                typeread += 1
                        elif typeread == 2:
                            if len(token) > 0:
                                self.addInit(line)
                                typeread += 1
                            else:
                                typeread += 1
                        elif typeread == 3:
                            if len(token) > 0:
                                self.addDst(line)
                                typeread += 1
                            else:
                                raise Exception('RAMSES: At least one dstfile is necessary in the command file.')
                                return
                        elif typeread == 4:
                            if len(token) > 0:
                                self.addTrj(line)
                                typeread += 1
                            else:
                                typeread += 2
                        elif typeread == 5:
                            if len(token) > 0:
                                self.addObs(line)
                                typeread += 1
                            else:
                                raise Exception(
                                    'RAMSES: Since a trajectory file was defined, an observable file should be given.')
                        elif typeread == 6:
                            if len(token) > 0:
                                self.addCont(line)
                                typeread += 1
                            else:
                                typeread += 1
                        elif typeread == 7:
                            if len(token) > 0:
                                self.addDisc(line)
                                typeread += 1
                            else:
                                typeread += 1
                        elif typeread == 8:
                            if len(token) > 0:
                                self.addRunObs(line)
                            else:
                                typeread += 1
            except IOError as e:
                raise IOError("RAMSES: I/O error({0}): {1}".format(e.errno, e.strerror))

    def writeCmdFile(self, userFile=None):
        """Write command file.

        :param str userFile: The filename to write to (Optional). The complete path can be given or
                             a the path relative to the working directory (os.getcwd()). If a file is not given, then a
                             temporary is created and deleted when the object is destroyed.

        """
        if not self._dataset or not self._dstset:
            raise RAMSESError('Dataset and disturbance set cannot be empty.')

        cmdFile = ""

        try:
            for i in self._dataset:
                cmdFile = cmdFile + i + '\n'
            cmdFile = cmdFile + '\n'
            if self._init:
                cmdFile = cmdFile + self._init[0] + '\n'
            else:
                cmdFile = cmdFile + '\n'
            if self._dstset:
                cmdFile = cmdFile + self._dstset[0] + '\n'
            if self._trj:
                cmdFile = cmdFile + self._trj[0] + '\n'
                cmdFile = cmdFile + self._obs[0] + '\n'
            else:
                cmdFile = cmdFile + '\n'
            if self._cont:
                cmdFile = cmdFile + self._cont[0] + '\n'
            else:
                cmdFile = cmdFile + '\n'
            if self._disc:
                cmdFile = cmdFile + self._disc[0] + '\n'
            else:
                cmdFile = cmdFile + '\n'
            for i in self._runobs:
                cmdFile = cmdFile + i + '\n'
            cmdFile = cmdFile + '\n'
            cmdFile = cmdFile + '\n'
        except IOError as e:
            raise IOError("RAMSES: I/O error({0}): {1}".format(e.errno, e.strerror))

        if userFile == None:
            return cmdFile
        else:
            if os.path.isfile(userFile):
                warnings.warn('The file %s already exists. It will be overwritten!' % (userFile))
            text_file = open(userFile, "w")
            text_file.write(cmdFile)
            text_file.close()
            return 0

    # def __del__(self):
        # silentremove(self._cmdfile.name)

    def addInit(self, afile):
        """Define the file where the simulation initialization will be saved.
        
        The initialization file is where the simulator will write the initialization procedure output.

        :param str afile: the filename. The complete path can be given or a the path relative to the working directory (os.getcwd())

        :Example:

        >>> import pyramses
        >>> case1 = pyramses.cfg()
        >>> case1.addInit("init.trace")

        .. warning:: If the file already exists, it will be ovewritten without warning!
        """
        del self._init[:]
        if os.path.isfile(afile):
            warnings.warn('The file %s already exists. It will be overwritten!' % (afile))
        self._init.append(afile)

    def getInit(self):
        """Return initialization trace file
        
        :returns: name of file
        :rtype: str
        
        """
        if self._init:
            return self._init[0]
        else:
            return ''

    def addOut(self, afile):
        """Define the file where the simulation output will be saved.

        :param str afile: the filename. The complete path must be given.

        :Example:

        >>> import pyramses
        >>> import os
        >>> case1 = pyramses.cfg()
        >>> case1.addOut(os.path.join(os.getcwd(),'output.trace'))

        .. note:: If the file already exists, the output will be appended to that file.
        """
        del self._out[:]
        self._out.append(afile)

    def getOut(self):
        """Return output trace file
        
        :returns: name of file
        :rtype: str
        
        """
        if self._out:
            return self._out[0]
        else:
            return ''

    def addCont(self, afile):
        """Add contrinuous trace file
        
        The continuous trace file saves information about the convergence of the solution algorithm
        used inside RAMSES. This is mainly used for debugging reasons and it can slow down the execution
        of the simulation.

        :param str afile: the filename. The complete path can be given or a the path relative to the working directory (os.getcwd())

        :Example:

        >>> import pyramses
        >>> case1 = pyramses.cfg()
        >>> case1.addCont("cont.trace")

        .. warning:: If the file already exists, it will be ovewritten without warning!
        """
        del self._cont[:]
        if os.path.isfile(afile):
            warnings.warn('The file %s already exists. It will be overwritten!' % (afile))
        self._cont.append(afile)

    def getCont(self):
        """Return contrinuous trace file
        
        :returns: name of file
        :rtype: str
        
        """
        if self._cont:
            return self._cont[0]
        else:
            return ''

    def addTrj(self, afile):
        """Add trajectory file

        :param str afile: the filename. The complete path can be given or a the path relative to the working directory (os.getcwd())

        :Example:

        >>> import pyramses
        >>> case1 = pyramses.cfg()
        >>> case1.addTrj("output.trj")

        .. warning:: If the file already exists, it will be ovewritten without warning!
        """
        del self._trj[:]
        if os.path.isfile(afile):
            warnings.warn('The file %s already exists. It will be overwritten!' % (afile))
        self._trj.append(afile)

    def getTrj(self):
        """Return trajectory file
        
        :returns: name of file
        :rtype: str
        
        """
        if self._trj:
            return self._trj[0]
        else:
            return ''

    def addDisc(self, afile):
        """Add discrete trace file.
        
        The discrete trace file saves information about the discrete events in the system, these may be
        from the discrete controllers, events in the disturbance file, or from discrete variables inside
        the injector, twoport, torque, or exciter models.

        :param str afile: the filename. The complete path can be given or a the path relative to the working directory (os.getcwd())

        :Example:

        >>> import pyramses
        >>> case1 = pyramses.cfg()
        >>> case1.addDisc("disc.trace")

        .. warning:: If the file already exists, it will be ovewritten without warning!
        """
        del self._disc[:]
        if os.path.isfile(afile):
            warnings.warn('The file %s already exists. It will be overwritten!' % (afile))
        self._disc.append(afile)

    def clearDisc(self):
        """Clear discrete trace file

        """
        del self._disc[:]

    def getDisc(self):
        """Return discrete trace file
        
        :returns: name of discrete trace file
        :rtype: str
        
        """
        if self._disc:
            return self._disc[0]
        else:
            return ''

    def addRunObs(self, runobs):
        """Add runtime observable.
        
        This defines some states that will be displayed during the simulation using gnuplot.
        
        :param str runobs: Description of runtime observable as defined in the RAMSES userguide
        
        :Example:        
        
        - BV BUSNAME: Voltage magnitude of bus::
   
           >>> case.addRunObs('BV 1041') # observe the bus voltage of bus 1041
        
        - MS SYNHRONOUS_MACHINE: Synchronous speed of machine::
        
           >>> case.addRunObs('MS g1')
           
        - RT RT: Real-time vs simulated time plot::
        
           >>> case.addRunObs('RT RT')
           
        - BPE/BQE/BPO/BQO BRANCH_NAME: Branch active (P), reactive (Q) power at the origin (O) or extremity (E) of a branch::
        
           >>> case.addRunObs('BPE 1041-01') # active power at the origin of branch 1041-01
        
        - ON INJECTOR_NAME OBSERVABLE_NAME: Monitor a named observable from an injector ::

           >>> case.addRunObs('ON WT1a Pw') # observable Pw from injector WT1a


        """
        if __runTimeObs__:
            if runobs not in self._runobs:
                self._runobs.append(runobs)
        else:
            warnings.warn('Gnuplot is not the path, so runtime observables are disabled.')

    def clearRunObs(self):
        """Clear runtime observables

        """
        del self._runobs[:]

    def addObs(self, ofile):
        """Add observables file

        :param str ofile: the filename. The complete path can be given or a the path relative to the working directory (os.getcwd())

        :Example:

        >>> import pyramses
        >>> case1 = pyramses.cfg()
        >>> case1.addObs("obs.dat")

        """
        if os.path.isfile(ofile):
            del self._obs[:]
            self._obs.append(ofile)
        else:
            raise IOError('RAMSES: The observables file %s does not exist or is not valid.' % (ofile))

    def getObs(self):
        """Get observables file name
        
        :returns: name of observables file
        :rtype: str
        
        """
        if self._obs:
            return self._obs[0]
        else:
            return ''

    def addData(self, datafile):
        """Add datafile in the dataset.
        
        The data files provide a description of the system to be simulated, along with the solver parameters.

        :param str datafile: the filename. The complete path can be given or a the path relative to the working directory (os.getcwd())

        :Example:

        >>> import pyramses
        >>> case1 = pyramses.cfg()
        >>> case1.addData("dyn_A.dat")

        .. warning:: At least one file needs to be added.
        """
        if os.path.isfile(datafile):
            if datafile not in self._dataset:
                self._dataset.append(datafile)
        else:
            raise IOError('RAMSES: The datafile %s does not exist or is not valid.' % (datafile))

    def getData(self):
        """Get datafile list
        
        :returns: list of data file names
        :rtype: list
        
        """
        return self._dataset

    def delData(self, datafile):
        """Remove a specific datafile from dataset
        
        """
        if datafile in self._dataset:
            self._dataset.remove(datafile)

    def clearData(self):
        """Empty the dataset
        
        """
        del self._dataset[:]

    def addDst(self, dstfile):
        """Add dstfile in the dstset
        
        The disturbance file is where the disturbance to be simulated is described

        :param str dstfile: the filename. The complete path can be given or a the path relative to the working directory (os.getcwd())

        :Example:

        >>> import pyramses
        >>> case1 = pyramses.cfg()
        >>> case1.addDst("short.dst")

        .. warning:: A file needs to be provided for the simulation to start.
        """
        if os.path.isfile(dstfile):
            del self._dstset[:]
            self._dstset.append(dstfile)
        else:
            raise IOError('RAMSES: The dstfile %s does not exist or is not valid.' % (dstfile))

    def getDst(self):
        """Get disturbance
        
        :returns: name of disturbance file
        :rtype: str
        
        """
        if self._dstset:
            return self._dstset[0]
        else:
            return ''

    def clearDst(self):
        """Empty the dstset
        
        """
        del self._dstset[:]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Global module variables and subroutines for pyramses."""

import errno
import inspect
import os

__runTimeObs__ = True
__libdir__ = os.path.realpath(
    os.path.abspath(os.path.join(os.path.split(inspect.getfile(inspect.currentframe()))[0], "libs")))


def CustomWarning(message, category, filename, lineno, file=None, line=None):
    """Custom routine to print warnings."""
    print("RAMSESWarning: %s" % message)


def read_file(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


class RAMSESError(Exception):
    pass


def __which(program):
    """Check the path for a program and if it's executable."""

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def silentremove(filename):
    """Check if a file exists and delete it
    
    :type filename: str
    """

    try:
        os.remove(filename)
    except OSError as e:
        if e.errno != errno.ENOENT:  # errno.ENOENT = no such file or directory
            raise  # re-raise exception if a different error occured

def wrapToList(item):
    """Wraps item into a list if it's not already a list.
    """
    if not isinstance(item, list):
        return [item]
    else:
        return item

    


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""PyRAMSES module."""

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages
import os

def read_file(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

install_requires = ['matplotlib','scipy','numpy','mkl==2024.2.2']

from pyramses import __version__, __author__, __email__, __status__, __url__, __name__

setup(
    name=__name__,
    version=__version__,
    description='Python library for RAMSES dynamic simulator.',
    author=__author__,
    author_email=__email__,
    url=__url__,
    keywords=['RAMSES', 'Power Systems', 'Simulator'],
    license='Proprietary',
    long_description=read_file('README.rst'),
    long_description_content_type='text/x-rst',
    packages=find_packages(),
    install_requires=install_requires, 
    package_data={
        'pyramses': ['libs/*.dll', 'libs/*.so', 'libs/*.h'],
    },
    classifiers=[
        "Development Status :: " + __status__,
        "Intended Audience :: Developers",
        "Environment :: Console",
        "License :: Other/Proprietary License",
        "Programming Language :: Python :: 3"
    ],
    entry_points={
        'console_scripts' : [
            'dynsim = pyramses.scripts.exec:run',
        ]
    }

)
